@@ -385,14 +385,17 @@ public:
 		StringKeyTable<Entry>::Iterator p_it(storeWithPreviewData.entries);
 		StringKeyTable<Entry>::ConstIterator it(entries);
 		vector<Error> tmpErrors;
-		Error error;
 
 		while (*p_it != NULL) {
 			const HashedStaticString &key = p_it.getKey();
 			Entry &entry = p_it.getValue();
 
 			if (isWritable(entry) && updates.isMember(key)) {
-				entry.userValue = updates[key];
+				bool ok = entry.schemaEntry->tryTypecastValue(
+					updates[key], entry.userValue);
+				if (!ok) {
+					entry.userValue = updates[key];
+				}
 			}
 
 			p_it.next();
@@ -406,7 +409,11 @@ public:
 			entry.schemaEntry->inspect(subdoc);
 
 			if (isWritable(entry) && updates.isMember(key)) {
-				subdoc["user_value"] = updates[key];
+				bool ok = entry.schemaEntry->tryTypecastValue(updates[key],
+					subdoc["user_value"]);
+				if (!ok) {
+					subdoc["user_value"] = updates[key];
+				}
 			} else {
 				subdoc["user_value"] = entry.userValue;
 			}
@@ -419,9 +426,7 @@ public:
 				subdoc["effective_value"] =
 					getEffectiveValue(subdoc["user_value"],
 						subdoc["default_value"]);
-			if (!schema->validateValue(it.getKey(), effectiveValue, error)) {
-				tmpErrors.push_back(error);
-			}
+			schema->validateValue(it.getKey(), effectiveValue, tmpErrors);
 
 			result[it.getKey()] = subdoc;
 			it.next();
@@ -469,8 +474,7 @@ public:
 				if (isWritable(entry)) {
 					const Json::Value &subdoc =
 						const_cast<const Json::Value &>(preview)[it.getKey()];
-					entry.userValue = entry.schemaEntry->typecastValue(
-						subdoc["user_value"]);
+					entry.userValue = subdoc["user_value"];
 				}
 				it.next();
 			}
